@@ -9,7 +9,7 @@ pub use self::traversals::*;
 
 use crate::{
     DataId, ElementId, FunctionId, GlobalId, LocalFunction, MemoryId, ModuleTypes, RefType,
-    TableId, TypeId, ValType,
+    TableId, TagId, TypeId, ValType,
 };
 use id_arena::Id;
 use std::fmt;
@@ -219,6 +219,18 @@ pub(crate) enum BlockKind {
     /// An `Else` block
     Else,
 
+    /// A `try_table` block
+    TryTable,
+
+    /// A `try` block (legacy exception handling)
+    Try,
+
+    /// A `catch` handler block (legacy exception handling)
+    Catch,
+
+    /// A `catch_all` handler block (legacy exception handling)
+    CatchAll,
+
     /// The entry to a function.
     FunctionEntry,
 }
@@ -551,6 +563,33 @@ pub enum Instr {
         func: FunctionId,
     },
 
+    /// `ref.as_non_null` - assert reference is non-null or trap
+    RefAsNonNull {},
+
+    /// `br_on_null` - branch if reference is null
+    BrOnNull {
+        /// The block to branch to if the reference is null
+        block: InstrSeqId,
+    },
+
+    /// `br_on_non_null` - branch if reference is non-null
+    BrOnNonNull {
+        /// The block to branch to if the reference is non-null
+        block: InstrSeqId,
+    },
+
+    /// `call_ref` - call through a typed function reference
+    CallRef {
+        /// The type index of the function being called
+        ty: TypeId,
+    },
+
+    /// `return_call_ref` - tail call through a typed function reference
+    ReturnCallRef {
+        /// The type index of the function being called
+        ty: TypeId,
+    },
+
     /// `v128.bitselect`
     V128Bitselect {},
 
@@ -612,11 +651,100 @@ pub enum Instr {
         /// The table which `func` below is indexing into
         table: TableId,
     },
+
+    /// `try_table ... end` - exception handling with catch table (modern proposal)
+    #[walrus(skip_builder)]
+    TryTable {
+        /// The id of this `try_table` instruction's inner `InstrSeq`.
+        seq: InstrSeqId,
+        /// The catch clauses for this try block.
+        #[walrus(skip_visit)]
+        catches: Vec<TryTableCatch>,
+    },
+
+    /// `throw` - throw an exception
+    Throw {
+        /// The tag of the exception being thrown.
+        tag: TagId,
+    },
+
+    /// `throw_ref` - rethrow a caught exception reference
+    ThrowRef {},
+
+    // Legacy exception handling instructions (phase 1 proposal)
+    // Support these as long as browsers also support them
+    /// `try blocktype ... end` - try block with catch handlers (legacy)
+    #[walrus(skip_builder)]
+    Try {
+        /// The id of this `try` block's inner `InstrSeq` (the try body).
+        seq: InstrSeqId,
+        /// The catch handlers for this try block
+        #[walrus(skip_visit)]
+        catches: Vec<LegacyCatch>,
+    },
+
+    /// `rethrow relative_depth` - rethrow a caught exception (legacy)
+    Rethrow {
+        /// The relative depth of the catch block
+        #[walrus(skip_visit)]
+        relative_depth: u32,
+    },
 }
 
 /// Argument in `V128Shuffle` of lane indices to select
 pub type ShuffleIndices = [u8; 16];
 
+/// A catch clause in a legacy `Try` instruction
+#[derive(Clone, Debug)]
+pub enum LegacyCatch {
+    /// `catch tag` - catches exception with specific tag
+    Catch {
+        /// The tag to match
+        tag: TagId,
+        /// The catch handler block
+        handler: InstrSeqId,
+    },
+    /// `catch_all` - catches any exception
+    CatchAll {
+        /// The catch-all handler block
+        handler: InstrSeqId,
+    },
+    /// `delegate relative_depth` - delegates to outer block instead of catching
+    Delegate {
+        /// The relative depth to delegate to
+        relative_depth: u32,
+    },
+}
+
+/// A catch clause in a `TryTable` instruction
+#[derive(Clone, Debug)]
+pub enum TryTableCatch {
+    /// `catch tag label` - catches exception with specific tag, binds payload values
+    Catch {
+        /// The tag to match
+        tag: TagId,
+        /// The block to branch to
+        label: InstrSeqId,
+    },
+    /// `catch_ref tag label` - catches exception with specific tag, binds payload and exnref
+    CatchRef {
+        /// The tag to match
+        tag: TagId,
+        /// The block to branch to
+        label: InstrSeqId,
+    },
+    /// `catch_all label` - catches any exception
+    CatchAll {
+        /// The block to branch to
+        label: InstrSeqId,
+    },
+    /// `catch_all_ref label` - catches any exception, binds exnref
+    CatchAllRef {
+        /// The block to branch to
+        label: InstrSeqId,
+    },
+}
+
 /// Constant values that can show up in WebAssembly
 #[derive(Debug, Clone, Copy)]
 pub enum Value {
@@ -1269,7 +1397,11 @@ impl Instr {
             | Instr::BrTable(..)
             | Instr::Return(..)
             | Instr::ReturnCall(..)
-            | Instr::ReturnCallIndirect(..) => true,
+            | Instr::ReturnCallIndirect(..)
+            | Instr::ReturnCallRef(..)
+            | Instr::Throw(..)
+            | Instr::ThrowRef(..)
+            | Instr::Rethrow(..) => true,
 
             // No `_` arm to make sure that we properly update this function as
             // we add support for new instructions.
@@ -1309,6 +1441,10 @@ impl Instr {
             | Instr::RefNull(..)
             | Instr::RefIsNull(..)
             | Instr::RefFunc(..)
+            | Instr::RefAsNonNull(..)
+            | Instr::BrOnNull(..)
+            | Instr::BrOnNonNull(..)
+            | Instr::CallRef(..)
             | Instr::V128Bitselect(..)
             | Instr::I8x16Swizzle(..)
             | Instr::I8x16Shuffle(..)
@@ -1317,7 +1453,9 @@ impl Instr {
             | Instr::TableInit(..)
             | Instr::TableCopy(..)
             | Instr::ElemDrop(..)
-            | Instr::Drop(..) => false,
+            | Instr::Drop(..)
+            | Instr::TryTable(..)
+            | Instr::Try(..) => false,
         }
     }
 }
@@ -1359,3 +1497,59 @@ impl VisitMut for InstrSeq {
         }
     }
 }
+
+impl<'instr> Visit<'instr> for TryTableCatch {
+    fn visit<V>(&self, visitor: &mut V)
+    where
+        V: Visitor<'instr>,
+    {
+        match self {
+            TryTableCatch::Catch { tag, .. } | TryTableCatch::CatchRef { tag, .. } => {
+                visitor.visit_tag_id(tag);
+            }
+            TryTableCatch::CatchAll { .. } | TryTableCatch::CatchAllRef { .. } => {}
+        }
+    }
+}
+
+impl VisitMut for TryTableCatch {
+    fn visit_mut<V>(&mut self, visitor: &mut V)
+    where
+        V: VisitorMut,
+    {
+        match self {
+            TryTableCatch::Catch { tag, .. } | TryTableCatch::CatchRef { tag, .. } => {
+                visitor.visit_tag_id_mut(tag);
+            }
+            TryTableCatch::CatchAll { .. } | TryTableCatch::CatchAllRef { .. } => {}
+        }
+    }
+}
+
+impl<'instr> Visit<'instr> for LegacyCatch {
+    fn visit<V>(&self, visitor: &mut V)
+    where
+        V: Visitor<'instr>,
+    {
+        match self {
+            LegacyCatch::Catch { tag, .. } => {
+                visitor.visit_tag_id(tag);
+            }
+            LegacyCatch::CatchAll { .. } | LegacyCatch::Delegate { .. } => {}
+        }
+    }
+}
+
+impl VisitMut for LegacyCatch {
+    fn visit_mut<V>(&mut self, visitor: &mut V)
+    where
+        V: VisitorMut,
+    {
+        match self {
+            LegacyCatch::Catch { tag, .. } => {
+                visitor.visit_tag_id_mut(tag);
+            }
+            LegacyCatch::CatchAll { .. } | LegacyCatch::Delegate { .. } => {}
+        }
+    }
+}
