@@ -45,9 +45,6 @@ pub struct ModuleTranslation<'data> {
     /// Module information.
     pub module: Module,
 
-    /// This module's index.
-    pub module_index: StaticModuleIndex,
-
     /// The input wasm binary.
     ///
     /// This can be useful, for example, when modules are parsed from a
@@ -115,8 +112,7 @@ impl<'data> ModuleTranslation<'data> {
     /// Create a new translation for the module with the given index.
     pub fn new(module_index: StaticModuleIndex) -> Self {
         Self {
-            module_index,
-            module: Module::default(),
+            module: Module::new(module_index),
             wasm: &[],
             function_body_inputs: PrimaryMap::default(),
             known_imported_functions: SecondaryMap::default(),
@@ -139,6 +135,11 @@ impl<'data> ModuleTranslation<'data> {
             .as_ref()
             .expect("module type information to be available")
     }
+
+    /// Get this translation's module's index.
+    pub fn module_index(&self) -> StaticModuleIndex {
+        self.module.module_index
+    }
 }
 
 /// Contains function data: byte code and its offset in the module.
