@@ -719,13 +719,28 @@ pub enum Trampoline {
 
     /// A `resource.new` intrinsic which will inject a new resource into the
     /// table specified.
-    ResourceNew(TypeResourceTableIndex),
+    ResourceNew {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+        /// The type of the resource.
+        ty: TypeResourceTableIndex,
+    },
 
     /// Same as `ResourceNew`, but for the `resource.rep` intrinsic.
-    ResourceRep(TypeResourceTableIndex),
+    ResourceRep {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+        /// The type of the resource.
+        ty: TypeResourceTableIndex,
+    },
 
     /// Same as `ResourceNew`, but for the `resource.drop` intrinsic.
-    ResourceDrop(TypeResourceTableIndex),
+    ResourceDrop {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+        /// The type of the resource.
+        ty: TypeResourceTableIndex,
+    },
 
     /// A `backpressure.set` intrinsic, which tells the host to enable or
     /// disable backpressure for the caller's instance.
@@ -750,9 +765,10 @@ pub enum Trampoline {
     /// lifted export function.  This allows the callee to continue executing
     /// after returning a result.
     TaskReturn {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// Tuple representing the result types this intrinsic accepts.
         results: TypeTupleIndex,
-
         /// The canonical ABI options specified for this intrinsic.
         options: OptionsIndex,
     },
@@ -775,6 +791,8 @@ pub enum Trampoline {
     /// outstanding async task/stream/future to make progress, returning the
     /// first such event.
     WaitableSetWait {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// Configuration options for this intrinsic call.
         options: OptionsIndex,
     },
@@ -783,6 +801,8 @@ pub enum Trampoline {
     /// async task/stream/future has made progress.  Unlike `task.wait`, this
     /// does not block and may return nothing if no such event has occurred.
     WaitableSetPoll {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// Configuration options for this intrinsic call.
         options: OptionsIndex,
     },
@@ -802,7 +822,10 @@ pub enum Trampoline {
     /// A `thread.yield` intrinsic, which yields control to the host so that other
     /// tasks are able to make progress, if any.
     ThreadYield {
-        /// If `true`, indicates the caller instance maybe reentered.
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+        /// If `true`, indicates the caller instance maybe receive notification
+        /// of task cancellation.
         cancellable: bool,
     },
 
@@ -824,15 +847,18 @@ pub enum Trampoline {
     /// A `stream.new` intrinsic to create a new `stream` handle of the
     /// specified type.
     StreamNew {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `stream` type and caller instance.
         ty: TypeStreamTableIndex,
     },
 
     /// A `stream.read` intrinsic to read from a `stream` of the specified type.
     StreamRead {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `stream` type and caller instance.
         ty: TypeStreamTableIndex,
-
         /// Any options (e.g. string encoding) to use when storing values to
         /// memory.
         options: OptionsIndex,
@@ -840,9 +866,10 @@ pub enum Trampoline {
 
     /// A `stream.write` intrinsic to write to a `stream` of the specified type.
     StreamWrite {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `stream` type and caller instance.
         ty: TypeStreamTableIndex,
-
         /// Any options (e.g. string encoding) to use when storing values to
         /// memory.
         options: OptionsIndex,
@@ -851,6 +878,8 @@ pub enum Trampoline {
     /// A `stream.cancel-read` intrinsic to cancel an in-progress read from a
     /// `stream` of the specified type.
     StreamCancelRead {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `stream` type and caller instance.
         ty: TypeStreamTableIndex,
         /// If `false`, block until cancellation completes rather than return
@@ -861,6 +890,8 @@ pub enum Trampoline {
     /// A `stream.cancel-write` intrinsic to cancel an in-progress write from a
     /// `stream` of the specified type.
     StreamCancelWrite {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `stream` type and caller instance.
         ty: TypeStreamTableIndex,
         /// If `false`, block until cancellation completes rather than return
@@ -871,6 +902,8 @@ pub enum Trampoline {
     /// A `stream.drop-readable` intrinsic to drop the readable end of a
     /// `stream` of the specified type.
     StreamDropReadable {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `stream` type and caller instance.
         ty: TypeStreamTableIndex,
     },
@@ -878,6 +911,8 @@ pub enum Trampoline {
     /// A `stream.drop-writable` intrinsic to drop the writable end of a
     /// `stream` of the specified type.
     StreamDropWritable {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `stream` type and caller instance.
         ty: TypeStreamTableIndex,
     },
@@ -885,15 +920,18 @@ pub enum Trampoline {
     /// A `future.new` intrinsic to create a new `future` handle of the
     /// specified type.
     FutureNew {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `future` type and caller instance.
         ty: TypeFutureTableIndex,
     },
 
     /// A `future.read` intrinsic to read from a `future` of the specified type.
     FutureRead {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `future` type and caller instance.
         ty: TypeFutureTableIndex,
-
         /// Any options (e.g. string encoding) to use when storing values to
         /// memory.
         options: OptionsIndex,
@@ -901,9 +939,10 @@ pub enum Trampoline {
 
     /// A `future.write` intrinsic to write to a `future` of the specified type.
     FutureWrite {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `future` type and caller instance.
         ty: TypeFutureTableIndex,
-
         /// Any options (e.g. string encoding) to use when storing values to
         /// memory.
         options: OptionsIndex,
@@ -912,6 +951,8 @@ pub enum Trampoline {
     /// A `future.cancel-read` intrinsic to cancel an in-progress read from a
     /// `future` of the specified type.
     FutureCancelRead {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `future` type and caller instance.
         ty: TypeFutureTableIndex,
         /// If `false`, block until cancellation completes rather than return
@@ -922,6 +963,8 @@ pub enum Trampoline {
     /// A `future.cancel-write` intrinsic to cancel an in-progress write from a
     /// `future` of the specified type.
     FutureCancelWrite {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `future` type and caller instance.
         ty: TypeFutureTableIndex,
         /// If `false`, block until cancellation completes rather than return
@@ -932,6 +975,8 @@ pub enum Trampoline {
     /// A `future.drop-readable` intrinsic to drop the readable end of a
     /// `future` of the specified type.
     FutureDropReadable {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `future` type and caller instance.
         ty: TypeFutureTableIndex,
     },
@@ -939,6 +984,8 @@ pub enum Trampoline {
     /// A `future.drop-writable` intrinsic to drop the writable end of a
     /// `future` of the specified type.
     FutureDropWritable {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the specific `future` type and caller instance.
         ty: TypeFutureTableIndex,
     },
@@ -946,6 +993,8 @@ pub enum Trampoline {
     /// A `error-context.new` intrinsic to create a new `error-context` with a
     /// specified debug message.
     ErrorContextNew {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the `error-context` type in the caller instance.
         ty: TypeComponentLocalErrorContextTableIndex,
         /// String encoding, memory, etc. to use when loading debug message.
@@ -958,6 +1007,8 @@ pub enum Trampoline {
     /// Note that the debug message might not necessarily match what was passed
     /// to `error.new`.
     ErrorContextDebugMessage {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the `error-context` type in the caller instance.
         ty: TypeComponentLocalErrorContextTableIndex,
         /// String encoding, memory, etc. to use when storing debug message.
@@ -966,6 +1017,8 @@ pub enum Trampoline {
 
     /// A `error-context.drop` intrinsic to drop a specified `error-context`.
     ErrorContextDrop {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
         /// The table index for the `error-context` type in the caller instance.
         ty: TypeComponentLocalErrorContextTableIndex,
     },
@@ -1014,7 +1067,6 @@ pub enum Trampoline {
     AsyncStartCall {
         /// The callee's callback, if any.
         callback: Option<RuntimeCallbackIndex>,
-
         /// The callee's post-return function, if any.
         post_return: Option<RuntimePostReturnIndex>,
     },
@@ -1047,13 +1099,23 @@ pub enum Trampoline {
     ///
     /// The payload here represents that this is accessing the Nth slot of local
     /// storage.
-    ContextGet(u32),
+    ContextGet {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+        /// Which slot to access.
+        slot: u32,
+    },
 
     /// Intrinsic used to implement the `context.set` component model builtin.
     ///
     /// The payload here represents that this is accessing the Nth slot of local
     /// storage.
-    ContextSet(u32),
+    ContextSet {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+        /// Which slot to update.
+        slot: u32,
+    },
 }
 
 impl Trampoline {
@@ -1074,9 +1136,9 @@ impl Trampoline {
                 format!("component-transcode-{op}-m{from}-m{to}")
             }
             AlwaysTrap => format!("component-always-trap"),
-            ResourceNew(i) => format!("component-resource-new[{}]", i.as_u32()),
-            ResourceRep(i) => format!("component-resource-rep[{}]", i.as_u32()),
-            ResourceDrop(i) => format!("component-resource-drop[{}]", i.as_u32()),
+            ResourceNew { ty, .. } => format!("component-resource-new[{}]", ty.as_u32()),
+            ResourceRep { ty, .. } => format!("component-resource-rep[{}]", ty.as_u32()),
+            ResourceDrop { ty, .. } => format!("component-resource-drop[{}]", ty.as_u32()),
             BackpressureSet { .. } => format!("backpressure-set"),
             BackpressureInc { .. } => format!("backpressure-inc"),
             BackpressureDec { .. } => format!("backpressure-dec"),
@@ -1117,8 +1179,8 @@ impl Trampoline {
             FutureTransfer => format!("future-transfer"),
             StreamTransfer => format!("stream-transfer"),
             ErrorContextTransfer => format!("error-context-transfer"),
-            ContextGet(_) => format!("context-get"),
-            ContextSet(_) => format!("context-set"),
+            ContextGet { .. } => format!("context-get"),
+            ContextSet { .. } => format!("context-set"),
         }
     }
 }
