@@ -498,8 +498,10 @@ impl AddressTransform {
         >,
         wasm_file: wasmtime_environ::WasmFileInfo,
     ) -> Self {
+        use cranelift_entity::EntityRef;
+
         let mut translations = wasmtime_environ::PrimaryMap::new();
-        let mut translation = wasmtime_environ::ModuleTranslation::default();
+        let mut translation = wasmtime_environ::ModuleTranslation::new(StaticModuleIndex::new(0));
         translation.debuginfo.wasm_file = wasm_file;
         translation
             .module
@@ -552,8 +554,10 @@ impl AddressTransform {
     }
 
     fn translate_raw(&self, addr: u64) -> Option<(usize, GeneratedAddress)> {
-        if addr == 0 {
-            // It's normally 0 for debug info without the linked code.
+        const TOMBSTONE: u64 = u32::MAX as u64;
+        if addr == 0 || addr == TOMBSTONE {
+            // Addresses for unlinked code may be left as 0 or replaced
+            // with -1, depending on the linker used.
             return None;
         }
         if let Some(func) = self.find_func(addr) {
